#include <u.h>
#include <errno.h>
#include <stdio.h>
#include <libelf.h>
#include <libdwarf.h>
#include <dwarf.h>

#include "queue.h"
#include "dat.h"
#include "hash.h"
#include "type.h"
#include "cqctbridge.h"

/* Internal */
static	char*	cidentifier(char *name);
static	int	dwarfdiewalk(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, DwarfStack*);
static	int	dwarfdiewalk1(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, DwarfStack*);
static	int	dwarfdodie(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, Dwarf_Half, DwarfStack*, DwarfType*);
static	int	dwarfdodieattrs(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, Dwarf_Attribute*, int, DwarfType*, Type*);
static	int	dwarfdodiearrayattrs(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, Dwarf_Attribute*, int, DwarfType*, Type*);
static	int	dwarfdodieenumdeclattrs(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, Dwarf_Attribute*, int, DwarfType*, Decl*);
static	int	dwarfdodiememberdeclattrs(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, Dwarf_Attribute*, int, DwarfType*, Decl*);
static	int	dwarfdodieformalsdeclattrs(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, Dwarf_Attribute*, int, DwarfType*, Decl*);
static	int	dwarfdodievariabledeclattrs(Dwarf_Debug, Dwarf_Die, DwarfCompileUnit*, Dwarf_Attribute*, int, DwarfType*, DwarfVar*);

static	int	evalblock(Dwarf_Debug, DwarfCompileUnit*, DwarfAttrValue*, uintptr*, const char*);

extern Type* dwbasetypes [Vnbase];
extern void reset_dwbasetypes (void);
extern Type* get_dwbasetype (enum Cbase);

int
dwarfinfowalk(Dwarf_Debug dbg)
{
	int rv;
	Dwarf_Die cudie;
	Dwarf_Unsigned off;
	DwarfStack *stk;
	DwarfCompileUnit *cu;

	off = 0;
	bzero(&cudie, sizeof(cudie));

	for(;;){
		cu = mkdwarfcompileunit();
		cu->offset = off;
		rv = dwarf_next_cu_header(dbg, &cu->hdrlen, &cu->version,
				&cu->abbrevoff, &cu->addrsz,
				&cu->nextoff, &dwarferror);
		if(rv != DW_DLV_OK)
			break;
		rv = dwarf_siblingof(dbg, nil, &cudie, &dwarferror);
		if(rv == DW_DLV_ERROR)
			goto Error;
		else if(rv == DW_DLV_NO_ENTRY)
			goto Next;
		assert(rv == DW_DLV_OK);
		rv = dwarf_srcfiles(cudie, &cu->srcs, &cu->nsrcs, &dwarferror);
		if(rv != DW_DLV_OK){
			cu->srcs = nil;
			cu->nsrcs = 0;
		}
		rv = dwarf_srclines(cudie, &cu->lines, &cu->nlines, &dwarferror);
		if(rv != DW_DLV_OK){
			cu->srcs = nil;
			cu->nsrcs = 0;
		}
		stk = mkdwarfstack();
		dwarfpush(stk);
		rv = dwarfdiewalk(dbg, cudie, cu, stk);
		dwarfpop(stk);
		dwarffreestack(stk);
		dwarf_dealloc(dbg, cudie, DW_DLA_DIE);
		if(rv < 0)
			return -1;
	Next:
		off = cu->nextoff;
	}
	if(rv == DW_DLV_ERROR){
  Error:
		fprintf(stderr, "error: %s\n", dwarf_errmsg(dwarferror));
		return -1;
	}
	return 0;
}

/* return a copy of name if it is a valid C identifier, otherwise a
   new valid (but not necessarily unique) identifier based on name.
*/
static char*
cidentifier(char *name)
{
	char *s, *p, *q, c;
	
	s = malloc(strlen(name)+2); /* 1 extra in case leading non-alpha */
	p = name;
	q = s;
	c = name[0];
	if(!isalpha(c) && c != '_')
		*q++ = '_';
	while(c = *p){
		if(!isalpha(c) && !isdigit(c) && c != '_')
			*q = '_';
		else
			*q = c;
		p++;
		q++;
	}
	*q = 0;
	return s;
}

static int
dwarfdiewalk(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, DwarfStack *stk)
{
	int rv;
	Dwarf_Die child;
	Dwarf_Die sibling;
	Dwarf_Error error;

	/* Walk the DIE tree depth-first pre-order */
	rv = dwarfdiewalk1(dbg, die, cu, stk);
	if(rv < 0)
		return rv;
	rv = dwarf_child(die, &child, &error);
	if(rv == DW_DLV_OK){
		dwarfpush(stk);
		rv = dwarfdiewalk(dbg, child, cu, stk);
		dwarfpop(stk);
		dwarf_dealloc(dbg, child, DW_DLA_DIE);
		if(rv < 0)
			goto Error;
	}else if(rv == DW_DLV_ERROR){
		fprintf(stderr, "error: %s\n", dwarf_errmsg(error));
		goto Error;
	}
	rv = dwarf_siblingof(dbg, die, &sibling, &error);
	if(rv == DW_DLV_OK){	/* XXX: loop to eliminate recursion */
		rv = dwarfdiewalk(dbg, sibling, cu, stk);
		dwarf_dealloc(dbg, sibling, DW_DLA_DIE);
		if(rv < 0)
			goto Error;
	}else if(rv == DW_DLV_ERROR){
		fprintf(stderr, "error: %s\n", dwarf_errmsg(error));
		goto Error;
	}
	return 0;

  Error:
	return -1;
}

static int
dwarfdiewalk1(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, DwarfStack *stk)
{
	int rv;
	DwarfName *nm;
	Dwarf_Half tag;
	Dwarf_Off off, totoff;
	DwarfType *dtype;

	dtype = nil;

	rv = dwarf_tag(die, &tag, &dwarferror);
	if(rv != DW_DLV_OK){
		fprintf(stderr, "error: dwarf_tag: %s\n",
			dwarf_errmsg(dwarferror));
		goto Error;
	}

	if(verbose > 2){
		nm = dwarfname(tag);
		if(nm == nil)
			fprintf(stderr, "error: unknown tag %d\n", tag);
		else
			fprintf(stderr, "tag %d --> %s\n", tag, nm->name);
	}

	rv = dwarf_dieoffset(die, &totoff, &dwarferror);
	if(rv != DW_DLV_OK){
		fprintf(stderr, "error: dwarf_dieoffset: %s\n",
			dwarf_errmsg(dwarferror));
		goto Error;
	}
	rv = dwarf_die_CU_offset(die, &off, &dwarferror);
	if(rv != DW_DLV_OK){
		fprintf(stderr, "error: dwarf_die_CU_offset: %s\n",
			dwarf_errmsg(dwarferror));
		goto Error;
	}

	dtype = dwarftypefindlocal(cu->offset, off);
	if(dtype == nil)
		dtype = dwarftypefindglobal(totoff);
	if(dtype != nil)
		goto Done;
	dtype = mkdwarftype(cu, tag, totoff, off);
	rv = dwarfdodie(dbg, die, cu, tag, stk, dtype);
	if(rv < 0)
		goto Error;
  Done:
	dwarfsettop(stk, dtype);
	return tag;

  Error:
	return -1;
}

static int
dwarfdodie(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, Dwarf_Half tag, DwarfStack *stk, DwarfType *dtype)
{
	int rv;
	int error;
	Dwarf_Signed na, i;
	Dwarf_Attribute *alist;
	DwarfType *parent;

	Type *t;
	Decl *d;
	DwarfVar *v;

	error = 0;
	alist = nil;

	rv = dwarf_attrlist(die, &alist, &na, &dwarferror);
	if(rv == DW_DLV_ERROR){
		fprintf(stderr, "error: dwarf_attrlist: %s\n",
			dwarf_errmsg(dwarferror));
		goto Error;
	}else if(rv == DW_DLV_NO_ENTRY)
		na = 0;

	switch(tag){
	case DW_TAG_typedef:
		t = mktype(Ttypedef);
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		/*
		 * Typedef's are decls if DW_AT_type is missing.
		 * If DWARF were well-designed, it would require
	         * DW_AT_declaration here, too.
                 */
		if(t->dwarf.addr == 0 && t->dwarf.offset == 0)
			t->flags |= Fdecl;
		dtype->type = t;
		break;

	case DW_TAG_subrange_type:
		/* XXX: assume C order */
		parent = dwarffindparent(stk, Tarr);
		t = parent->type;
		rv = dwarfdodiearrayattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		break;

	case DW_TAG_array_type:
		/* XXX: fill in dimension */
		t = mktype(Tarr);
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		dtype->type = t;
		break;

	case DW_TAG_enumeration_type:
		t = mktype(Tenum);
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		dtype->type = t;
		break;

	case DW_TAG_structure_type:
		t = mktype(Tstruct);
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		dtype->type = t;
		break;

	case DW_TAG_union_type:
		t = mktype(Tunion);
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		dtype->type = t;
		break;

	case DW_TAG_base_type: {
	  Expr *e;
	  t = mktype(Tbase);
	  rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
	  if(rv < 0)
	    goto Error;
	  e = isstdbasetype (t->tag);
	  if (e == NULL) {
	    /* then we need a typdef to tie this base type to a standard one*/
	    int encoding;
	    Type *tdt;
	    enum DW_bytesizes size;
	    tdt = mktype (Ttypedef);
	    tdt->tag = strdup (t->tag);
	    tdt->size = t->size;
	    if (t->size == 0) {
	      if (strcmp (t->tag, "VOID"))
		tdt->encoding = Vvoid;
	    }
	    else {
	      size = mapsizes (t->size);
	      encoding = t->encoding;
	      if (encoding == 128) {
		tdt->encoding = Vfloat;
	      }
	      else {
		tdt->encoding = stdbasetype [size][encoding];
	      }
	    }
	    dtype->type = tdt;
	  }
	  else
	    dtype->type = t;
	}
	  break;

	case DW_TAG_enumerator:
		d = mkdecl();
		rv = dwarfdodieenumdeclattrs(dbg, die, cu, alist, na, dtype, d);
		if(rv < 0)
			goto Error;
		parent = dwarffindsueparent(stk);
		if(parent == nil)
			goto Error;
		STAILQ_INSERT_TAIL(&parent->type->members, d, link);
		break;

	case DW_TAG_member:
		d = mkdecl();
		rv = dwarfdodiememberdeclattrs(dbg, die, cu, alist, na, dtype, d);
		if(rv < 0)
			goto Error;
		parent = dwarffindsueparent(stk);
		if(parent == nil)
			goto Error;
		STAILQ_INSERT_TAIL(&parent->type->members, d, link);
		break;

	case DW_TAG_const_type:
	case DW_TAG_pointer_type:
	case DW_TAG_reference_type:
	case DW_TAG_restrict_type:
	case DW_TAG_volatile_type:
		t = nil;
		switch(tag){
		case DW_TAG_const_type:
			t = mktype(Tconst);
			break;

		case DW_TAG_pointer_type:
			t = mktype(Tptr);
			break;

		case DW_TAG_reference_type:
			t = mktype(Tref);
			break;

		case DW_TAG_restrict_type:
			t = mktype(Trestrict);
			break;

		case DW_TAG_volatile_type:
			t = mktype(Tvolatile);
			break;
		}
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		dtype->type = t;
		break;

	case DW_TAG_subprogram:
		d = mkdecl();
		v = mkdwarfvar(d, 0);
		rv = dwarfdodievariabledeclattrs(dbg, die, cu, alist, na, dtype, v);
		if(rv < 0)
			goto Error;
		/* Do a little dance to make type resolution work */
		t = mktype(Tfun);
		t->dwarf = d->dwarf;
		d->dwarf.addr = dtype->addr;
		d->dwarf.offset = dtype->offset;
		d->dwarf.compile = dtype->compile;
		d->type = t;
		dtype->type = t;
		break;

	case DW_TAG_inlined_subroutine:
		/* XXX: handle abstract origin crud */
		t = mktype(Tfun);
		t->flags |= Finline;
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		dtype->type = t;
		break;

	case DW_TAG_subroutine_type:
		t = mktype(Tfun);
		rv = dwarfdodieattrs(dbg, die, cu, alist, na, dtype, t);
		if(rv < 0)
			goto Error;
		dtype->type = t;
		break;

	case DW_TAG_formal_parameter:
		parent = dwarffindparent(stk, Tfun);
		/* XXX: appear to be some bogus cases where this happens? */
		if(parent == nil)
			goto Done;
		d = mkdecl();
		rv = dwarfdodieformalsdeclattrs(dbg, die, cu, alist, na, dtype, d);
		if(rv < 0)
			goto Error;
		STAILQ_INSERT_TAIL(&parent->type->members, d, link);
		break;

	case DW_TAG_unspecified_parameters:
		d = mkdecl();
		d->dwarf.addr = ~0;		/* XXX: magic */
		parent = dwarffindparent(stk, Tfun);
		if(parent == nil)
			goto Error;
		STAILQ_INSERT_TAIL(&parent->type->members, d, link);
		break;

	case DW_TAG_variable:
		/* Suppress locals */
		parent = dwarffindparent(stk, Tfun);
		if(parent != nil)
			goto Done;
		d = mkdecl();
		v = mkdwarfvar(d, 0);
		rv = dwarfdodievariabledeclattrs(dbg, die, cu, alist, na, dtype, v);
		if(rv < 0)
			goto Error;
		break;

	case DW_TAG_class_type:
	case DW_TAG_string_type:
	case DW_TAG_constant:
		break;

	default:
		/* NADA */
		break;
	}

  Done:
	if(alist){
		for(i=0; i<na; i++)
			dwarf_dealloc(dbg, alist[i], DW_DLA_ATTR);
		dwarf_dealloc(dbg, alist, DW_DLA_LIST);
	}
	if(error)
		return -1;
	return 0;

  Error:
	/* XXX: leak Type, Decl on error */
	error = 1;
	goto Done;
}

static int
dwarfdodiearrayattrs(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, Dwarf_Attribute *alist, int na, DwarfType *dt, Type *t)
{
	int i;
	int rv;
	Dwarf_Half attr;
	DwarfAttrValue val;
	u64int low, high;

	low = 0;
	high = 0;
	bzero(&val, sizeof val);
	for(i=0; i<na; i++){
		rv = dwarf_whatattr(alist[i], &attr, &dwarferror);
		if(rv != DW_DLV_OK)	/* Just skip it */
			continue;
		bzero(&val, sizeof val);
		rv = dwarfattrvalue(dbg, die, alist[i], attr, &val, cu);
		if(rv < 0)
			goto Error;
		switch(attr){
		case DW_AT_lower_bound:
			if(val.issigned)
				low =  (u64int)val.sdata;
			else
				low = val.udata;
			break;

		case DW_AT_upper_bound:
			if(val.issigned)
				high =  (u64int)val.sdata;
			else
				high = val.udata;
			break;

		case DW_AT_count:
			low = 0;
			if(val.issigned)
				high =  (u64int)val.sdata;
			else
				high = val.udata;
			break;
		}
	}

	dwarfaddarraydim(t, low, high);
	dwarfattrvalueclear(dbg, &val);
	return 0;

  Error:
	dwarfattrvalueclear(dbg, &val);
	return -1;
}

/*
 * Handle standard type attributes, including:
 *
 *	Attribute	Type field
 *	name		tag
 *	type		dwarfaddr, dwarfoff
 */
static int
dwarfdodieattrs(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, Dwarf_Attribute *alist, int na, DwarfType *dt, Type *t)
{
	int i;
	int rv;
	Dwarf_Half attr;
	Dwarf_Unsigned x;
	DwarfAttrValue val;

	bzero(&val, sizeof val);
	for(i=0; i<na; i++){
		rv = dwarf_whatattr(alist[i], &attr, &dwarferror);
		if(rv != DW_DLV_OK)	/* Just skip it */
			continue;
		bzero(&val, sizeof val);
		rv = dwarfattrvalue(dbg, die, alist[i], attr, &val, cu);
		if(rv < 0)
			goto Error;
		switch(attr){
		case DW_AT_name:
			t->tag = strdup(val.str);
			break;

		case DW_AT_type:
			t->dwarf.compile = cu;
			if(val.form == DW_FORM_addr)
				t->dwarf.addr = val.addr;
			else
				t->dwarf.offset = val.off;
			break;

		case DW_AT_byte_size:
			/* XXX: should be small enough that this is OK */
			if(val.issigned)
				t->size = val.sdata;
			else
				t->size = val.udata;
			break;

		case DW_AT_declaration:
			t->flags |= Fdecl;
			break;

		case DW_AT_decl_file:
			if(!val.isvalid)
				goto Error;
			x = val.udata;
			if(x > 0 && x-1 < cu->nlines){
				rv = dwarf_line_srcfileno(cu->lines[x-1],
					&val.udata, &dwarferror);
				if(rv != DW_DLV_OK)
					goto Error;
				x = val.udata;
				if(x > 0 && x < cu->nsrcs)
					dt->file = strdup(cu->srcs[x-1]);
			}
			break;

		case DW_AT_decl_line:
			if(!val.isvalid)
				goto Error;
			x = val.udata;
			if(x > 0 && x-1 < cu->nlines){
				rv = dwarf_lineno(cu->lines[x-1],
					&val.udata, &dwarferror);
				if(rv != DW_DLV_OK)
					goto Error;
				dt->line = val.udata;
			}
			break;
		case DW_AT_encoding:
		  if (!val.isvalid)
		    goto Error;
		  t->encoding = val.udata;
		  break;
		}
	}

	if(t->tag == nil)
		t->tag = anontype();

	dwarfattrvalueclear(dbg, &val);
	return 0;

  Error:
	dwarfattrvalueclear(dbg, &val);
	return -1;
}

/*
 * Handle standard enum decl member attributes, including:
 *
 *	Attribute	Type field
 */
static int
dwarfdodieenumdeclattrs(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, Dwarf_Attribute *alist, int na, DwarfType *dt, Decl *d)
{
	int i;
	int rv;
	uintptr xval;
	Dwarf_Half attr;
	DwarfAttrValue val;

	bzero(&val, sizeof val);
	for(i=0; i<na; i++){
		rv = dwarf_whatattr(alist[i], &attr, &dwarferror);
		if(rv != DW_DLV_OK)	/* Just skip it */
			continue;
		bzero(&val, sizeof val);
		rv = dwarfattrvalue(dbg, die, alist[i], attr, &val, cu);
		if(rv < 0)
			goto Error;
		switch(attr){
		case DW_AT_name:
			d->name = cidentifier(val.str);
			break;

		case DW_AT_const_value:
			switch(val.form){
			case DW_FORM_block:
			case DW_FORM_block1:
			case DW_FORM_block2:
			case DW_FORM_block4:
				xval = 0;
				rv = dwarfevalconst(dbg, cu, val.block->bl_data,
					val.block->bl_len, &xval);
				if(rv < 0)
					goto Error;
				dwarfattrvalueclear(dbg, &val);
				val.isvalid = 1;
				val.form = DW_FORM_udata;
				val.udata = (Dwarf_Off)xval;
				break;
	
			case DW_FORM_data1:
			case DW_FORM_data2:
			case DW_FORM_data4:
			case DW_FORM_data8:
			case DW_FORM_sdata:
			case DW_FORM_string:
			case DW_FORM_strp:
			case DW_FORM_udata:
				break;

			default:
				fprintf(stderr, "buggered const_value\n");
				goto Error;
			}
			/* Pass value to decl */
			memmove(&d->value, &val, sizeof val);
			memset(&val, 0, sizeof val);
			break;

		case DW_AT_declaration:
			abort();
			break;
		}
	}

	if(d->name == nil)
		d->name = anondecl();

	dwarfattrvalueclear(dbg, &val);
	return 0;

  Error:
	dwarfattrvalueclear(dbg, &val);
	return -1;
}

/*
 * Handle standard struct/union decl member attributes, including:
 *
 *	Attribute	Type field
 */
static int
dwarfdodiememberdeclattrs(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, Dwarf_Attribute *alist, int na, DwarfType *dt, Decl *d)
{
	int i;
	int rv;
	int dobits;
	uintptr xval;
	Dwarf_Half attr;
	DwarfAttrValue val;

	dobits = 0;
	bzero(&val, sizeof val);
	for(i=0; i<na; i++){
		rv = dwarf_whatattr(alist[i], &attr, &dwarferror);
		if(rv != DW_DLV_OK)	/* Just skip it */
			continue;
		bzero(&val, sizeof val);
		rv = dwarfattrvalue(dbg, die, alist[i], attr, &val, cu);
		if(rv < 0)
			goto Error;
		switch(attr){
		case DW_AT_name:
			d->name = cidentifier(val.str);
			break;

		case DW_AT_type:
			d->dwarf.compile = cu;
			if(val.form == DW_FORM_addr)
				d->dwarf.addr = val.addr;
			else
				d->dwarf.offset = val.off;
			break;

		case DW_AT_data_member_location:
			rv = evalblock(dbg, cu, &val, &xval, "data_member_location");
			if(rv < 0)
				goto Error;
			d->offset = (int)xval;
			break;

		case DW_AT_byte_size:		/* bit field: container size */
			dobits = 1;
			rv = evalblock(dbg, cu, &val, &xval, "byte_size");
			if(rv < 0)
				goto Error;
			d->bytesize = (int)xval;
			break;

		case DW_AT_bit_size:		/* bit field: number of bits */
			dobits = 1;
			rv = evalblock(dbg, cu, &val, &xval, "bit_size");
			if(rv < 0)
				goto Error;
			d->bitsize = (int)xval;
			break;

		case DW_AT_bit_offset:		/* bit field: bits to left */
			dobits = 1;
			rv = evalblock(dbg, cu, &val, &xval, "bit_offset");
			if(rv < 0)
				goto Error;
			d->msboffset = (int)xval;
			break;

		case DW_AT_declaration:
			abort();
			break;
		}
	}

	if(dobits){
		switch(endianness){
		case BigEndian:
			d->bitoffset = d->msboffset;
			break;

		case LittleEndian:
			if(d->bytesize == 0){
				fprintf(stderr, "size of integral type containing bit field not specified\n");
				goto Error;
			}
			d->bitoffset =
				8*d->bytesize - d->msboffset - d->bitsize;
			break;

		default:
			abort();
		}
	}

	if(d->name == nil)
		d->name = anondecl();

	dwarfattrvalueclear(dbg, &val);
	return 0;

  Error:
	dwarfattrvalueclear(dbg, &val);
	return -1;
}

static int
dwarfdodieformalsdeclattrs(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, Dwarf_Attribute *alist, int na, DwarfType *dt, Decl *d)
{
	int i;
	int rv;
	Dwarf_Half attr;
	DwarfAttrValue val;

	bzero(&val, sizeof val);
	for(i=0; i<na; i++){
		rv = dwarf_whatattr(alist[i], &attr, &dwarferror);
		if(rv != DW_DLV_OK)	/* Just skip it */
			continue;
		bzero(&val, sizeof val);
		rv = dwarfattrvalue(dbg, die, alist[i], attr, &val, cu);
		if(rv < 0)
			goto Error;
		switch(attr){
		case DW_AT_name:
			d->name = cidentifier(val.str);
			break;

		case DW_AT_type:
			d->dwarf.compile = cu;
			if(val.form == DW_FORM_addr)
				d->dwarf.addr = val.addr;
			else
				d->dwarf.offset = val.off;
			break;
		}
	}

	dwarfattrvalueclear(dbg, &val);
	return 0;

  Error:
	dwarfattrvalueclear(dbg, &val);
	return -1;
}

/*
 * Handle both variables and functions (DWARF subprograms).
 */
static int
dwarfdodievariabledeclattrs(Dwarf_Debug dbg, Dwarf_Die die, DwarfCompileUnit *cu, Dwarf_Attribute *alist, int na, DwarfType *dt, DwarfVar *v)
{
	int i;
	int rv;
	uintptr xval;
	Dwarf_Half attr;
	DwarfAttrValue val;

	Decl *d;

	d = v->decl;
	bzero(&val, sizeof val);
	for(i=0; i<na; i++){
		rv = dwarf_whatattr(alist[i], &attr, &dwarferror);
		if(rv != DW_DLV_OK)	/* Just skip it */
			continue;
		bzero(&val, sizeof val);
		rv = dwarfattrvalue(dbg, die, alist[i], attr, &val, cu);
		if(rv < 0)
			goto Error;
		switch(attr){
		case DW_AT_low_pc:	/* Use low pc as location for funcs */
		case DW_AT_location:
			switch(val.form){
			case DW_FORM_addr:
				val.udata = val.addr;
				break;

			case DW_FORM_block:
			case DW_FORM_block1:
			case DW_FORM_block2:
			case DW_FORM_block4:
				xval = 0;
				rv = dwarfevalconst(dbg, cu, val.block->bl_data,
					val.block->bl_len, &xval);
				if(rv == BadOp){ }
				else if(rv < 0)
					goto Error;
				else{
					dwarfattrvalueclear(dbg, &val);
					val.isvalid = 1;
					val.form = DW_FORM_udata;
					val.udata = (Dwarf_Off)xval;
				}
				break;

			case DW_FORM_data1:
			case DW_FORM_data2:
			case DW_FORM_data4:
			case DW_FORM_data8:
			case DW_FORM_sdata:
			case DW_FORM_udata:
				break;

			case DW_FORM_string:
			case DW_FORM_strp:
				break;
			}
			if(val.isvalid)
				v->addr = (uintptr)val.udata;
			break;

		case DW_AT_name:
			d->name = cidentifier(val.str);
			break;

		case DW_AT_type:
			d->dwarf.compile = cu;
			if(val.form == DW_FORM_addr)
				d->dwarf.addr = val.addr;
			else
				d->dwarf.offset = val.off;
			break;
		}
	}

	dwarfattrvalueclear(dbg, &val);
	return 0;

  Error:
	dwarfattrvalueclear(dbg, &val);
	return -1;
}

/* Caller guarantees that val is valid (i.e. val->isvalid != 0) */
static int
evalblock(Dwarf_Debug dbg, DwarfCompileUnit *cu, DwarfAttrValue *val, uintptr *xval, const char *msg)
{
	int rv;

	*xval = 0;

	switch(val->form){
	case DW_FORM_block:
	case DW_FORM_block1:
	case DW_FORM_block2:
	case DW_FORM_block4:
		rv = dwarfevalconst(dbg, cu, val->block->bl_data,
			val->block->bl_len, xval);
		if(rv < 0)
			goto Error;
		break;

	case DW_FORM_data1:
	case DW_FORM_data2:
	case DW_FORM_data4:
	case DW_FORM_data8:
	case DW_FORM_sdata:
	case DW_FORM_udata:
		if(val->issigned)
			*xval = (int)val->sdata;
		else
			*xval = (int)val->udata;
		break;

	default:
		fprintf(stderr, "buggered %s: %s\n",
			msg, dwarfformname(val->form));
		fflush(stderr);
		goto Error;
	}
	return 0;

  Error:
	return -1;
}
