#include <u.h>
#include <stdio.h>
#include <dwarf.h>
#include <libdwarf.h>

#include "queue.h"
#include "dat.h"
#include "hash.h"
#include "type.h"
#include "util.h"
#include "cqctbridge.h"

void reset_dwbasetypes (void);
Type* get_dwbasetype (enum Cbase);
Type* dwbasetypes [Vnbase] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

typedef struct DTHash DTHash;
struct DTHash {
	int		ntab;
	DwarfTypeList	*tab;
};

/*
 * DWARF does not make void a base type the way it should.
 * As a result, we use global address zero and local address 0.0
 * to indicate a void type.  We do mark it as Tbase.
 *
 * We use a similar hack for variadic arguments, but use the
 * addresses ~0 and 0.~0, respectively.  Type, however, is Tvararg.
 *
 * Also note that old-style C function declarations have one
 * argument in DWARF which is ``unspecified arguments'', which we
 * ordinarily map to var args.  The printer knows about this special
 * case and removes it from the output accordingly.
 *
 * Function declarations that take void don't have any argument
 * types in DWARF.  We patch up that case below.
 */

DwarfType	*dwarfvoid;
DwarfType	*dwarfvararg;
DwarfTypeList	dwarftypes = TAILQ_HEAD_INITIALIZER(dwarftypes);
DTHash		*dthashlocal;
DTHash		*dthashglobal;

char		*anonuniq;
static	uint	anonnum;

static	int	dwarfresolvetype(Type*);

static	u32int	dwarftypehashglobal(u64int);
static	u32int	dwarftypehashlocal(u64int, u32int);

static	DTHash		*mkdwarftypehash(void);
static	void		dwarftypeinsertglobal(DwarfType*);
static	void		dwarftypeinsertlocal(DwarfType*);
static	DwarfType	*dwarftypelookglobal(u64int);
static	DwarfType	*dwarftypelooklocal(u64int, u32int);
static void fixtype (Type *);
void fixbrokendwarfarraytypes (void);


static char*
mkanon(void)
{
	int n;
	char *s;

	anonnum++;
	n = 4+1+20+1+1;
	if(anonuniq){
		n += strlen(anonuniq);
		s = malloc(n);
		if(s == nil)
			goto Error;
		snprintf(s, n, "anon_%s_%u", anonuniq, anonnum);
	}else{
		s = malloc(n);
		if(s == nil)
			goto Error;
		snprintf(s, n, "anon_%u", anonnum);
	}
	return s;

  Error:
	return nil;
}

/*
 * Use same counter to reduce confusion --
 * may actually be more confusing, I suppose.
 */
char*
anontype(void)
{
	char *s;

	s = mkanon();
	return s;
}

char*
anondecl(void)
{
	char *s;

	s = mkanon();
	return s;
}

void
dwarfprinttypes(void)
{
	Type *t;
	DwarfType *dt;
	HT *ht;

	ht = mkht();
	TAILQ_FOREACH(dt, &dwarftypes, link){
		t = dt->type;
		if(t != nil && t->type == Ttypedef && !wasseen(ht, t->tag))
			printtype(t);
	}
	freeseen(ht);
	freeht(ht);

	ht = mkht();
	TAILQ_FOREACH(dt, &dwarftypes, link){
		t = dt->type;
		if(t != nil && t->type != Ttypedef && (t->flags & Fdecl)!=0
		   && !wasseen(ht, t->tag))
			printtype(dt->type);
	}
	freeseen(ht);
	freeht(ht);

	ht = mkht();
	TAILQ_FOREACH(dt, &dwarftypes, link){
		t = dt->type;
		if(t != nil && t->type != Ttypedef && (t->flags & Fdecl)==0){
			if((t->type == Tstruct || t->type == Tunion
			    || t->type == Tenum)
			   && wasseen(ht, t->tag))
				continue;
			printtype(dt->type);
		}
	}
	freeseen(ht);
	freeht(ht);
}

Type*
mktype(int k)
{
	Type *ty;

	ty = malloc(sizeof *ty);
	memset(ty, 0, sizeof *ty);
	ty->type = k;
	STAILQ_INIT(&ty->members);
	return ty;
}

Decl*
mkdecl(void)
{
	Decl *d;

	d = malloc(sizeof *d);
	memset(d, 0, sizeof *d);
	return d;
}

int
dwarfresolvetypes(void)
{
	int rv;
	DwarfType *dt;

	TAILQ_FOREACH(dt, &dwarftypes, link){
		if(dt->type == nil)
			continue;
		rv = dwarfresolvetype(dt->type);
		if(rv < 0)
			return -1;
	}
	return 0;
}

static int
dwarfresolvetype(Type *t)
{
	int rv;
	Decl *d;
	DwarfType *dt;

//	if(t->flags & Fdecl)	/* Just a declaration: nothing to resolve */
//		return 0;
	if(t->flags & Finline)
		return 0;

	switch(t->type){
	case Tplastic:
	case Tbase:
	case Tenum:
		break;		/* Nothing to resolve */

	case Tfun:
		dt = type2dwarftype(t);
		if(dt != nil && t->sub == nil)
			t->sub = dt->type;
		if(dt->type == nil){
			fprintf(stderr, "unhandled Dwarf Type '%s'\n",
				dwarftagname(dt->tag)->name);
			dt->type = mktype(Tplastic);
		}
		if(t->sub == nil)
			t->sub = mktype(Tplastic);

		/* Resolve types for formals */
		STAILQ_FOREACH(d, &t->members, link){
			rv = dwarfresolvedecl(d);
			if(rv < 0)
				return -1;
		}

		/* Patch up functions with old and void declarations */
		d = STAILQ_FIRST(&t->members);
		if(d == nil){
			d = mkdecl();
			d->type = dwarfvoid->type;
			STAILQ_INSERT_HEAD(&t->members, d, link);
		}else if(d->type->type == Tvararg){
			if(STAILQ_NEXT(d, link) != nil)
				abort();
			STAILQ_REMOVE(&t->members, d, Decl, link);
		}
		break;

	case Tstruct:
	case Tunion:
		STAILQ_FOREACH(d, &t->members, link){
			rv = dwarfresolvedecl(d);
			if(rv < 0)
				return -1;
		}
		break;

	case Tarr:
	case Tptr:
	case Tref:
	case Tconst:
	case Tvolatile:
	case Trestrict:
	case Ttypedef:
	  if (t->encoding == 0) {
	    dt = type2dwarftype(t);
	    if(dt != nil && t->sub == nil)
	      t->sub = dt->type;
	    
	    if(dt->type == nil){
	      fprintf(stderr, "unhandled Dwarf Type '%s'\n",
		      dwarftagname(dt->tag)->name);
	      dt->type = mktype(Tplastic);
	    }
	    
	    if(t->sub == nil)
	      t->sub = mktype(Tplastic);
	  }
	  else
	    //t->sub = get_dwbasetype ((enum Cbase) (t->encoding));
	    t->sub = (dwarftypelookglobal (t->encoding))->type;
	  break;
	}
	return 0;
}

int
dwarfresolvedecl(Decl *d)
{
	DwarfType *dt;

	dt = decl2dwarftype(d);
	if(dt == nil)
		return -1;
	d->type = dt->type;
	return 0;
}

DwarfType*
mkdwarftype(DwarfCompileUnit *compile, ushort tag, u64int addr, u32int offset)
{
	DwarfType *dt;

	dt = malloc(sizeof *dt);
	memset(dt, 0, sizeof *dt);
	dt->tag = tag;
	dt->addr = addr;
	dt->offset = offset;
	dt->compile = compile;
	dwarftypeinsert(dt);
	return dt;
}

/*
 * DWARF is just another name for ``brain damaged''.
 * Cons up the void base type since it ought to be there but isn't.
 * DwarfRefs that aren't set via a DW_AT_type will be zero, so just
 * usurp that value for our purposes.
 */
void
dwarftypeinit(void)
{
	DwarfType *dt;
	DwarfCompileUnit *cu;
	Cbase cb;

#if 0
	cu = mkdwarfcompileunit();	/* Just leave blank */
	dt = malloc(sizeof *dt);
	memset(dt, 0, sizeof *dt);
	dt->tag = DW_TAG_base_type;	/* DWARF is borked */
	dt->addr = 0;
	dt->offset = 0;
	dt->compile = cu;
	dt->type = mktype(Tbase);
	dt->type->tag = strdup("void");
	dwarfvoid = dt;
#endif
	
	cu = mkdwarfcompileunit();	/* Just leave blank */
	dt = malloc(sizeof *dt);
	memset(dt, 0, sizeof *dt);
	dt->tag = DW_TAG_base_type;	/* DWARF is borked */
	dt->addr = ~0;
	dt->offset = ~0;
	dt->compile = cu;
	dt->type = mktype(Tvararg);
	dt->type->tag = strdup("...");
	dwarfvararg = dt;

	dthashlocal = mkdwarftypehash();
	dthashglobal = mkdwarftypehash();
	//dwarftypeinsertlocal(dwarfvoid);
	//dwarftypeinsertglobal(dwarfvoid);
	dwarftypeinsertlocal(dwarfvararg);
	dwarftypeinsertglobal(dwarfvararg);
	for (cb = Vchar; cb < Vnbase; cb++) {
	  cu = mkdwarfcompileunit();	/* Just leave blank */
	  dt = malloc(sizeof *dt);
	  memset(dt, 0, sizeof *dt);
	  dt->tag = DW_TAG_base_type;
	  dt->addr = cb; /* fix later; for now assume no addresses this low*/
	  dt->offset = 0;
	  dt->compile = cu;
	  dt->type = mktype (Tbase);
	  dt->type->tag = strdup (cbname [cb]);
	  dwarftypeinsertlocal (dt);
	  dwarftypeinsertglobal (dt);
	  if (cb == Vvoid)
	    dwarfvoid = dt;
	}
	return;
}

void
dwarftypeinsert(DwarfType *dt)
{
	dwarftypeinsertlocal(dt);
	dwarftypeinsertglobal(dt);
	TAILQ_INSERT_TAIL(&dwarftypes, dt, link);
	return;
}

DwarfType*
decl2dwarftype(Decl *d)
{
	DwarfType *dt;

	dt = dwarfvoid;
	if(d->dwarf.addr)
		dt = dwarftypefindglobal(d->dwarf.addr);
	else if(d->dwarf.offset)
		dt = dwarftypefindlocal(d->dwarf.compile->offset, d->dwarf.offset);
	return dt;
}

DwarfType*
type2dwarftype(Type *t)
{
	DwarfType *dt;

	dt = dwarfvoid;
	if(t->dwarf.addr)
		dt = dwarftypefindglobal(t->dwarf.addr);
	else if(t->dwarf.offset)
		dt = dwarftypefindlocal(t->dwarf.compile->offset, t->dwarf.offset);
	return dt;
}

DwarfType*
dwarftypefindglobal(u64int addr)
{
	DwarfType *dt;

	dt = dwarftypelookglobal(addr);
	return dt;
}

DwarfType*
dwarftypefindlocal(u64int unit, u32int offset)
{
	DwarfType *dt;

	dt = dwarftypelooklocal(unit, offset);
	return dt;
}

static u32int
dwarftypehashglobal(u64int addr)
{
	u32int h;

	h = 1677251;			/* Just a prime */
	h = hash64(addr, h);
	return h;
}

static u32int
dwarftypehashlocal(u64int unit, u32int offset)
{
	u32int h;

	h = 1677251;			/* Just a prime */
	h = hash64(unit, h);
	h = hash32(offset, h);
	return h;
}

static DTHash*
mkdwarftypehash(void)
{
	int i;
	DTHash *ht;

	ht = malloc(sizeof *ht);
	memset(ht, 0, sizeof *ht);
	ht->ntab = (1<<20)+7;
	ht->tab = malloc(ht->ntab*sizeof(ht->tab[0]));
	memset(ht->tab, 0, ht->ntab*sizeof(ht->tab[0]));
	for(i=0; i<ht->ntab; i++)
		TAILQ_INIT(&ht->tab[i]);
	return ht;
}

static void
dwarftypeinsertglobal(DwarfType *dt)
{
	u32int h;

	h = dwarftypehashglobal(dt->addr);
	h %= dthashglobal->ntab;
	TAILQ_INSERT_HEAD(&dthashglobal->tab[h], dt, hashglobal);
	return;
}

static void
dwarftypeinsertlocal(DwarfType *dt)
{
	u32int h;

	h = dwarftypehashlocal(dt->compile->offset, dt->offset);
	h %= dthashlocal->ntab;
	TAILQ_INSERT_HEAD(&dthashlocal->tab[h], dt, hashlocal);
	return;
}

static DwarfType*
dwarftypelookglobal(u64int addr)
{
	u32int h;
	DwarfType *dt;

	h = dwarftypehashglobal(addr);
	h %= dthashglobal->ntab;
	TAILQ_FOREACH(dt, &dthashglobal->tab[h], hashglobal)
		if(dt->addr == addr)
			return dt;
	return nil;
}

static DwarfType*
dwarftypelooklocal(u64int unit, u32int offset)
{
	u32int h;
	DwarfType *dt;

	h = dwarftypehashlocal(unit, offset);
	h %= dthashlocal->ntab;
	TAILQ_FOREACH(dt, &dthashlocal->tab[h], hashlocal)
		if(dt->compile->offset == unit && dt->offset == offset)
			return dt;
	return nil;
}

void
dwarfaddarraydim(Type *t, u64int low, u64int high)
{
	if(t->narr == t->nalloc){
		t->nalloc += 16;
		t->dim = realloc(t->dim, t->nalloc*sizeof(t->dim[0]));
	}
	t->dim[t->narr] = high-low + 1;
	t->narr++;
	return;
}

char*
typetab[] = {
	"plastic",
	"base",
	"struct",
	"union",
	"enum",
	"arr",
	"ptr",
	"typedef",
	"fun",
	"vararg",
	"ref",
	"const",
	"volatile",
	"restrict",
	nil
};

void
reset_dwbasetypes (void)
{
  enum Cbase cb;
  for (cb = Vundef; cb <= Vnbase; cb++)
    dwbasetypes [cb] = (Type *) NULL;
  return;
}

Type *
get_dwbasetype (enum Cbase cb)
{
  return dwbasetypes [cb];
}

void
fixbrokendwarfarraytypes (void)
{
  Type *t;
  DwarfType *dt;
  HT *ht;

  ht = mkht();
  TAILQ_FOREACH(dt, &dwarftypes, link){
    t = dt->type;
    if(t != nil && t->type != Ttypedef && (t->flags & Fdecl)==0){
      if((t->type == Tstruct || t->type == Tunion)
	 && wasseen(ht, t->tag))
	continue;
      fixtype(dt->type);
    }
  }
  freeseen(ht);
  freeht(ht);
}

static void
fixtype (Type * t)
{
  int ssize;
  Decl *d;

  ssize = t->size;
  STAILQ_FOREACH(d, &t->members, link) {
    int foff, nextoff;
    Decl *nextf;
    Type *ft;

    foff = d->offset;
    nextf = STAILQ_NEXT (d, link);
    if (nextf)
      nextoff = nextf->offset;
    else
      nextoff = -1;
    if (d->type && (foff == nextoff || foff == ssize)) {
      /* then fix the array type to have a first dimension of 0 */
      ft = d->type;
      if (ft->type == Tarr && ft->dim[0] == 1) {
	Type *newt;
	newt = (Type *) malloc (sizeof (Type));
	if (newt == 0)
	  return;
	memcpy ((void *) newt, (void *)ft, sizeof (Type));
	/* We should probably copy internal pointers like dim */
	newt->dim[0] = 0;
	d->type = newt;
      }
    }
  }
}
